# RUN AS ROOT
export BING_MAPS_API_KEY=YOUR_BING_KEY_HERE
export MAPQUEST_API_KEY=YOUR_MAPQUEST_KEY_HERE
export PELIAS_API_KEY=YOUR_PELIAS_KEY_HERE
export GOOGLE_API_KEY=YOUR_GOOGLE_KEY_HERE
export ESRI_CLIENT_ID=YOUR_ESRI_CLIENT_ID_HERE
export ESRI_CLIENT_SECRET=YOUR_ESRI_CLIENT_SECRET_HERE

python setup.py test $@
